/* -*- compile-command: "msdev modeler.dsp /make \"Modeler - Win32 Debug\"" -*-
 */

#include "modeler.h"

Tcl_HashTable htControls;
Tcl_Interp* theinterp;
Tcl_Obj* varArray;

/* Controls_Init
 *
 * called to initialize the control widgets package.
 *
 * controls work the following way:
 *
 *   - user calls, say, scale(), from his/her model setup function,
 *     giving the scale a unique integer control ID
 * 
 *   - the scale() function creates a new Tk widget in the controls
 *     panel which keeps its value in an element of the global hash
 *     Tcl variable "g_controls".  the hash key is the name of the Tk
 *     widget.
 *
 *   - this package keeps its own hash "htControls" which maps the
 *     integer control ID onto useful data about the control, such as the
 *     corresponding widget name.
 *
 *   - when user requests the control value by specifying the control
 *     ID, we look up the widget name in the htControls hash, then use
 *     that to look up the value in the g_controls hash, and return it.
 */

int Controls_Init( Tcl_Interp* interp )
{
    Tcl_InitHashTable( &htControls, TCL_ONE_WORD_KEYS );
    theinterp = interp;

    /* initialize the g_controls hash. */
    sprintf( command_buffer, "set g_controls(\"\") \"\"" );
    Tcl_GlobalEval( interp, command_buffer );

    /* keep a pointer to the name of the hash storing values. */
    varArray = Tcl_NewStringObj( "g_controls", -1 );
    Tcl_IncrRefCount( varArray );
    
    return TCL_OK;
}

/* new_control
 *
 * this does all the common functionality for creating a new control
 * of any type.
 */

control* new_control( int id, int type )
{
    char buffer[20];
    static int count = 1;
    control* p;
    Tcl_HashEntry* he;
    int newhe;

    /* allocate a new control structure.  initialize it with the
       control type (CONTROL_SCALE, CONTROL_CHECKBOX, etc.) */
    p = (control*)malloc( sizeof( control ) );
    p->type = type;
    p->data = NULL;

    /* add a entry to the htControls hash for this item. */
    he = Tcl_CreateHashEntry( &htControls, (char*)id, &newhe );
    if ( ! newhe )
    {
	/* if there already was a control with this ID, delete it. */
	
	control* oldp;
	oldp = (control*)Tcl_GetHashValue( he );
	sprintf( command_buffer, "destroy %s", Tcl_GetStringFromObj( oldp->widget, NULL ) );
	Tcl_Eval( theinterp, command_buffer );
	Tcl_DecrRefCount( p->widget );
	free( p );
    }

    /* the name of the Tk widget will be ".c.c#", where # is a unique
       number (NOT necessarily equal to the control ID, though!) */
    sprintf( buffer, ".c.c%d", count++ );
    /* store the name of the widget (as an object pointer) in the
       control structure. */
    p->widget = Tcl_NewStringObj( buffer, -1 );
    Tcl_IncrRefCount( p->widget );

    /* store the new entry in the htControls hash. */
    Tcl_SetHashValue( he, p );

    return p;
}

/* reset_controls
 *
 * deletes all control widgets and empties all the hashes.  used when
 * a model DLL is unloaded.
 */

void reset_controls( void )
{
    Tcl_HashSearch hs;
    Tcl_HashEntry* he;

    /* iterate he through every entry in the htControls hash. */
    he = Tcl_FirstHashEntry( &htControls, &hs );
    while ( he )
    {
	/* delete the Tk widget associated with this control. */
	control* p;
	p = (control*)Tcl_GetHashValue( he );
	sprintf( command_buffer, "destroy %s",
		 Tcl_GetStringFromObj( p->widget, NULL ) );
	Tcl_Eval( theinterp, command_buffer );
	Tcl_DecrRefCount( p->widget );
	
	/* free the control structure. */
	free( p->data );
	free( p );

	he = Tcl_NextHashEntry( &hs );
    }

    /* delete the htControls hash and reinitialize it (an easy way to
       empty it). */
    Tcl_DeleteHashTable( &htControls );
    Tcl_InitHashTable( &htControls, TCL_ONE_WORD_KEYS );

    /* delete the g_controls hash and reinitialize it. */
    sprintf( command_buffer, "unset g_controls; set g_controls(\"\") \"\"" );
    Tcl_GlobalEval( theinterp, command_buffer );

    /* update our pointer to the g_controls hash. */
    Tcl_DecrRefCount( varArray );
    varArray = Tcl_NewStringObj( "g_controls", -1 );
    Tcl_IncrRefCount( varArray );
}

void rebuild_controls( void )
{
    Tcl_HashSearch hs;
    Tcl_HashEntry* he;

    Debug( "call to rebuild_controls\n" );
    
    he = Tcl_FirstHashEntry( &htControls, &hs );
    while ( he )
    {
	control* p;
	p = (control*)Tcl_GetHashValue( he );
	switch( p->type )
	{
	    case CONTROL_SCALE:
		Debug( "rebuilding scale\n" );
		scale_widget( p );
		break;

	    case CONTROL_CURVE:
		curve_widget( p );
		break;

	    case CONTROL_CHECKBOX:
		checkbox_widget( p );
		break;

	    case CONTROL_RADIOBUTTON:
		radiobutton_widget( p );
		break;

	    case CONTROL_MENU:
		menu_widget( p );
		break;
	}

	he = Tcl_NextHashEntry( &hs );
    }
}

/* curve
 *
 * this function creates a curve control, which has a curve in the animation editor
 * but no widget in the control window.
 */

int curve( Tcl_Interp* interp, int id, char* name,
	   double min, double max, double start )
{
    control* p;
    curve_data* cd;
    
    /* create a new control. */
    p = new_control( id, CONTROL_CURVE );
    cd = (curve_data*)malloc( sizeof( curve_data ) );
    cd->start = start;
    p->data = cd;

    /* "name" is the label for the curve.  copy it to convert_buffer,
       escaping any spaces or funny characters. */
    CONVERT( name );

    /* tell curve animator about this curve. */
    sprintf( command_buffer, "add_curve %s %s %f %f %f",
	     Tcl_GetStringFromObj( p->widget, NULL ),
	     convert_buffer, min, max, start );
    Tcl_Eval( interp, command_buffer );

    curve_widget( p );
    
    return TCL_OK;
}

void curve_widget( control* p )
{
    curve_data* cd = (curve_data*)p->data;
    
    /* store the initial value "start" into the g_controls hash.  */
    Tcl_ObjSetVar2( theinterp, varArray, p->widget, Tcl_NewDoubleObj( cd->start ), TCL_GLOBAL_ONLY );
}
    
/* scale
 *
 * this function creates a slider control.  it should only be called
 * within the model setup function, unless you really know what you
 * are doing. 
 */

int scale( Tcl_Interp* interp, int id, char* name,
	   double min, double max, double res, double start )
{
    control* p;
    scale_data* sd;
    
    /* create a new control. */
    p = new_control( id, CONTROL_SCALE );

    sd = (scale_data*)malloc( sizeof( scale_data ) );
    p->data = sd;
    sd->min = min;
    sd->max = max;
    sd->res = res;
    sd->start = start;
    strncpy( sd->name, name, MAX_CONTROL_NAME );

    /* "name" is the label for the slider.  copy it to convert_buffer,
       escaping any spaces or funny characters. */
    CONVERT( name );

    /* tell curve animator about this slider. */
    sprintf( command_buffer, "add_curve %s %s %f %f %f",
	     Tcl_GetStringFromObj( p->widget, NULL ),
	     convert_buffer, min, max, start );
    Tcl_Eval( interp, command_buffer );

    scale_widget( p );
    
    return TCL_OK;
}

void scale_widget( control* p )
{
    scale_data* sd = (scale_data*)p->data;

    CONVERT( sd->name );
    
    /* create the Tk widget. */
    sprintf( command_buffer,
	     "scale %s -from %f -to %f -res %f -label %s -orient h -command control_change -variable g_controls(%s) -length 3i",
	     Tcl_GetStringFromObj( p->widget, NULL ), sd->min, sd->max, sd->res, convert_buffer,
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* pack the new widget into the control panel, causing it to
       appear onscreen. */
    sprintf( command_buffer, "pack %s -fill x -padx 2 -pady 2", Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );
    
    /* store the initial value "start" into the g_controls hash.  this
       will cause the slider on-screen to update automatically. */
    Tcl_ObjSetVar2( theinterp, varArray, p->widget, Tcl_NewDoubleObj( sd->start ), TCL_GLOBAL_ONLY );

}

/* checkbox
 *
 * this function creates a checkbox.  it should only be called
 * within the model setup function, unless you really know what you
 * are doing. 
 */

int checkbox( Tcl_Interp* interp, int id, char* name, int state )
{
    control* p;
    checkbox_data* cd;

    /* new control. */
    p = new_control( id, CONTROL_CHECKBOX );
    cd = (checkbox_data*)malloc( sizeof( checkbox_data ) );
    cd->state = state;
    strncpy( cd->name, name, MAX_CONTROL_NAME );
    p->data = cd;

    checkbox_widget( p );
    
    return TCL_OK;
}

void checkbox_widget( control* p )
{
    checkbox_data* cd = (checkbox_data*)p->data;

    /* escape characters within the name. */
    CONVERT( cd->name );

    Debug( "drawing checkbox %s\n", cd->name );
    
    /* create the widget. */
    sprintf( command_buffer,
	     "checkbutton %s -text %s -command control_change -variable g_controls(%s)",
	     Tcl_GetStringFromObj( p->widget, NULL ), convert_buffer,
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* set initial value. */
    Tcl_ObjSetVar2( theinterp, varArray, p->widget, Tcl_NewBooleanObj( cd->state ), TCL_GLOBAL_ONLY );

    /* pack the widget. */
    sprintf( command_buffer, "pack %s -fill x", Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );
}
    

/* radiobutton
 *
 * this function creates a set of radiobuttons.  it should only be
 * called within the model setup function, unless you really know what
 * you are doing.
 */

int radiobutton( Tcl_Interp* interp, int id, char* name, int start, ... )
{
    control* p;
    menu_data* md;
    va_list items;
    int i;
    char *item;

    /* new control. */
    p = new_control( id, CONTROL_RADIOBUTTON );
    md = (menu_data*)malloc( sizeof( menu_data ) );
    md->count = 0;
    md->start = start;
    strncpy( md->name, name, MAX_CONTROL_NAME );
    
    for ( i = 0, va_start( items, start ); item = va_arg( items, char* ); ++i )
    {
	strncpy( md->item[md->count], item, MAX_CONTROL_NAME );
	++md->count;
    }
    va_end( items );

    radiobutton_widget( p );
    
    return TCL_OK;
}

void radiobutton_widget( control* p )
{
    menu_data* md = (menu_data*)p->data;
    int i;

    /* the control widget will be a frame, which will contain the
       individual radio buttons as child widgets. */
    sprintf( command_buffer,
	     "frame %s -relief groove -bd 2",
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* put a label in the frame, labelling the whole group of
       choices. */
    CONVERT( md->name );
    sprintf( command_buffer,
	     "label %s.label -text %s -anchor w; pack %s.label -fill x",
	     Tcl_GetStringFromObj( p->widget, NULL ),
	     convert_buffer,
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* for each of the "..." arguments: */
    for ( i = 0; i < md->count; ++i )
    {
	/* escape funny characters in the item text. */
	CONVERT( md->item[i] );

	/* create a radiobutton widget for this option. */
	sprintf( command_buffer,
		 "radiobutton %s.r%d -text %s -value %d -variable g_controls(%s) -command control_change -anchor w; pack %s.r%d -fill x -padx 10",
		 Tcl_GetStringFromObj( p->widget, NULL ),
		 i, convert_buffer, i,
		 Tcl_GetStringFromObj( p->widget, NULL ),
		 Tcl_GetStringFromObj( p->widget, NULL ),
		 i );
	Tcl_Eval( theinterp, command_buffer );
    }

    /* set the initial choice. */
    Tcl_ObjSetVar2( theinterp, varArray, p->widget, Tcl_NewIntObj( md->start ), TCL_GLOBAL_ONLY );

    /* pack the frame with all the widgets inside it into the control
       panel. */
    sprintf( command_buffer, "pack %s -fill x -padx 2 -pady 2", Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );
}

/* menu
 *
 * this function creates menu with a set of radiobutton-style choices.
 * it works exactly like the "radiobutton" control; only the visual
 * appearance is different.  it should only be called within the model
 * setup function, unless you really know what you are doing.
 */

int menu( Tcl_Interp* interp, int id, char* name, int start, ... )
{
    control* p;
    menu_data* md;
    va_list items;
    int i;
    char* item;

    /* new control */
    p = new_control( id, CONTROL_MENU );
    md = (menu_data*)malloc( sizeof( menu_data ) );
    md->count = 0;
    md->start = start;
    strncpy( md->name, name, MAX_CONTROL_NAME );
    
    for ( i = 0, va_start( items, start ); item = va_arg( items, char* ); ++i )
    {
	strncpy( md->item[md->count], item, MAX_CONTROL_NAME );
	++md->count;
    }
    va_end( items );

    menu_widget( p );
    
    return TCL_OK;
}

void menu_widget( control* p )
{
    menu_data* md = (menu_data*)p->data;
    int i;
    
    /* create a frame to contain all the child widgets. */
    sprintf( command_buffer,
	     "frame %s",
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* put a label in the frame with the menu name. */
    CONVERT( md->name );
    sprintf( command_buffer,
	     "label %s.l -text %s -anchor e; pack %s.l -side left",
	     Tcl_GetStringFromObj( p->widget, NULL ),
	     convert_buffer,
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* create a menubutton, which will pop up the menu when pressed.
       the "text" field will get replaced. */
    sprintf( command_buffer,
	     "menubutton %s.mb -menu %s.mb.m -text hello -anchor w -relief raised; pack %s.mb -fill x -expand 1",
	     Tcl_GetStringFromObj( p->widget, NULL ),
	     Tcl_GetStringFromObj( p->widget, NULL ),
	     Tcl_GetStringFromObj( p->widget, NULL ),
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* create a menu. */
    sprintf( command_buffer,
	     "menu %s.mb.m -tearoff false",
	     Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );

    /* for each of the "..." arguments: */
    for ( i = 0; i < md->count; ++i )
    {
	/* escape funny characters. */
	CONVERT( md->item[i] );

	/* add an item to the menu with this option. */
	sprintf( command_buffer,
		 "%s.mb.m add radiobutton -label %s -value %d -variable g_controls(%s) -command {%s.mb config -text %s; control_change}",
		 Tcl_GetStringFromObj( p->widget, NULL ),
		 convert_buffer, i, 
		 Tcl_GetStringFromObj( p->widget, NULL ),
		 Tcl_GetStringFromObj( p->widget, NULL ),
		 convert_buffer );
	Tcl_Eval( theinterp, command_buffer );

	/* if this is the initial item, then set the menubutton's text
	   to this string. */
	if ( i == md->start )
	{
	    sprintf( command_buffer, "%s.mb config -text %s",
		     Tcl_GetStringFromObj( p->widget, NULL ),
		     convert_buffer );
	    Tcl_Eval( theinterp, command_buffer );
	}
    }

    /* put the initial value into the g_controls hash. */
    Tcl_ObjSetVar2( theinterp, varArray, p->widget, Tcl_NewIntObj( md->start ), TCL_GLOBAL_ONLY );

    /* pack the frame into the control panel. */
    sprintf( command_buffer, "pack %s -fill x -padx 2 -pady 2", Tcl_GetStringFromObj( p->widget, NULL ) );
    Tcl_Eval( theinterp, command_buffer );
}

/* get_control_d
 *
 * get the current value of a control, as a double.  can be called
 * from within any model function.
 */

double get_control_d( int id )
{
    Tcl_HashEntry* he;
    control* p;
    double value;

    /* look up the control structure. */
    he = Tcl_FindHashEntry( &htControls, (char*)id );
    /* if no control with this id, return 0. */
    if ( he == NULL )
	return 0.0;
    p = (control*)Tcl_GetHashValue( he );

    /* fetch the value from the g_controls hash. */
    if ( Tcl_GetDoubleFromObj( theinterp, Tcl_ObjGetVar2( theinterp, varArray, p->widget, TCL_GLOBAL_ONLY ), &value ) != TCL_OK )
	return 0.0;
    return value;
}

/* get_control_i
 *
 * get the current value of a control, as an int.  can be called
 * from within any model function.
 */

int get_control_i( int id )
{
    Tcl_HashEntry* he;
    control* p;
    int value;

    /* look up the control structure. */
    he = Tcl_FindHashEntry( &htControls, (char*)id );
    /* if no control with this ID, return 0. */
    if ( he == NULL )
	return 0;
    p = (control*)Tcl_GetHashValue( he );

    /* fetch the value from the g_controls hash. */
    if ( Tcl_GetIntFromObj( theinterp, Tcl_ObjGetVar2( theinterp, varArray, p->widget, TCL_GLOBAL_ONLY ), &value ) != TCL_OK )
	return 0;
    return value;
}

/* get_control_b
 *
 * get the current value of a control, as a boolean.  can be called
 * from within any model function.
 */

int get_control_b( int id )
{
    Tcl_HashEntry* he;
    control* p;
    int value;

    /* look up the control structure. */
    he = Tcl_FindHashEntry( &htControls, (char*)id );
    /* if no control with this ID, return 0 (false). */
    if ( he == NULL )
	return 0;
    p = (control*)Tcl_GetHashValue( he );

    /* fetch the value from the g_controls hash. */
    if ( Tcl_GetBooleanFromObj( theinterp, Tcl_ObjGetVar2( theinterp, varArray, p->widget, TCL_GLOBAL_ONLY ), &value ) != TCL_OK )
	return 0;
    return value;
}

