/* -*- compile-command: "msdev modeler.dsp /make \"Modeler - Win32 Debug\"" -*-
 */

#include <modeler.h>

/*
 * to add a new curve:
 *
 *  1. write a function whose prototype looks like this:
 *
 *        void generate( double xmin, double xmax, int wrapflag, double def,
 *                       int nc, double* xc, double* yc,
 *                       int* np, double** xp, double** yp );
 *
 *     the function will take a set of control points (in xc and yc, count in nc),
 *     generate a set of points (in xp and yp, count in np) describing the curve.
 *
 *     the returned set of points MUST have the following characteristics:
 *
 *       - the first point has x = xmin
 *
 *       - the last point has x = xmax
 *
 *       - points are sorted by x coordinate.
 *
 *     your function is responsible for allocating memory in *xp and *yp.
 *     MEMORY MUST BE ALLOCATED WITH "malloc()", NOT WITH "new"!!
 *
 *     if wrapflag is true, then the y value of the first point MUST equal the
 *     y value of the last point.
 *
 *     you must always return at least two points.  if nc = 0, return the two points
 *     (xmin,def) and (xmax,def).
 *
 *     see linear.cpp for a simple example.
 *
 *  2. name the curve type and add it to the curvetypes[] array below.
 *
 */

curvetype curvetypes[] = {
    { "linear", linear_generate },
    /* list additional curve types here, along with their generating functions. */

    /* leave NULL, NULL to mark the end of the list. */
    { NULL, NULL }
};

Tcl_HashTable htCurves;

int Curves_Init( Tcl_Interp* interp )
{
    Tcl_InitHashTable( &htCurves, TCL_STRING_KEYS );

    Tcl_CreateObjCommand( interp, "evaluate_curve", EvaluateCurveCmd,
			  NULL, NULL );
    Tcl_CreateObjCommand( interp, "get_evaluated_curve", GetEvaluatedCurveCmd,
			  NULL, NULL );
    Tcl_CreateObjCommand( interp, "set_curve_time", SetCurveTimeCmd,
			  NULL, NULL );
    Tcl_CreateObjCommand( interp, "get_curve_types", GetCurveTypesCmd,
			  NULL, NULL );
    
    return TCL_OK;
}

/* returns a list of available curve types (the names in the curvetypes[]
 * array).
 */
int GetCurveTypesCmd( ClientData clientdata, Tcl_Interp* interp,
		      int objc, Tcl_Obj* const objv[] )
{
    curvetype* ctp;
    
    if ( objc != 1 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"\"", TCL_STATIC );
	return TCL_ERROR;
    }

    for ( ctp = curvetypes; ctp->name != NULL; ++ctp )
	Tcl_ListObjAppendElement( interp, Tcl_GetObjResult( interp ), Tcl_NewStringObj( ctp->name, -1 ) );

    return TCL_OK;
}

/* SetCurveTimeCmd
 *
 * takes an animation time and an array name.  interpolates each curve to find its
 * value at that time, setting the corresponding element in the array to the value.
 */

int SetCurveTimeCmd( ClientData clientdata, Tcl_Interp* interp,
		     int objc, Tcl_Obj* const objv[] )
{
    double time;
    double value;
    Tcl_HashSearch hs;
    Tcl_HashEntry* he;

    if ( objc != 3 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"time arrayname\"", TCL_STATIC );
	return TCL_ERROR;
    }

    /* get the time. */
    if ( Tcl_GetDoubleFromObj( interp, objv[1], &time ) != TCL_OK )
	return TCL_ERROR;

    /* iterate over each curve. */
    he = Tcl_FirstHashEntry( &htCurves, &hs );
    while ( he )
    {
	/* find the value at the given time. */
	value = interpolate_curve_value( (evaluatedcurve*)Tcl_GetHashValue( he ), time );

	/* set the array element corresponding to the curve's name. */
	Tcl_ObjSetVar2( interp, objv[2], Tcl_NewStringObj( Tcl_GetHashKey( &htCurves, he ), -1 ), Tcl_NewDoubleObj( value ), 0 );
	
	he = Tcl_NextHashEntry( &hs );
    }

    return TCL_OK;
}

/* interpolate_curve_value
 *
 * uses binary search to locate the curve segment containing time t,
 * then does linear interpolation.
 */

double interpolate_curve_value( evaluatedcurve* ec, double t )
{
    int s, e, m;
    double f;

    if ( ec->n == 0 ) return ec->def;
    
    if ( t <= ec->x[0] ) return ec->y[0];
    if ( t >= ec->x[ec->n-1] ) return ec->y[ec->n-1];

    s = 0;
    e = ec->n - 1;

    while ( s < e-1 )
    {
	m = (s+e) / 2;
	if ( t < ec->x[m] )
	    e = m;
	else if ( t > ec->x[m] )
	    s = m;
	else
	    return ec->y[m];
    }

    f = ( t - ec->x[s] ) / ( ec->x[e] - ec->x[s] );
    return ec->y[s] + f * (ec->y[e] - ec->y[s]);
}

/* sortvar and index_compare
 *
 * used to sort an array of indices into sortvar.
 */

double* sortvar;

int index_compare( const void* a, const void* b )
{
    if ( sortvar[*(int*)a] < sortvar[*(int*)b] )
	return -1;
    else if ( sortvar[*(int*)a] > sortvar[*(int*)b] )
	return 1;
    else
	return 0;
}

/* EvaluateCurveCmd
 *
 * takes a bunch of stuff, including a set of control points, calls the curve
 * generate function, and saves the list of points in the hash where
 * SetCurveTimeCmd can get them.
 */

int EvaluateCurveCmd( ClientData clientdata, Tcl_Interp* interp, 
		      int objc, Tcl_Obj* const objv[] )
{
    int nc;
    double* xc;
    double* yc;
    double* tc;
    int* ic;
    
    Tcl_Obj** listv;
    Tcl_Obj* listel;
    int i;
    double xmin, xmax, def;
    int wrapflag;
    Tcl_HashEntry* he;
    int newhe;
    evaluatedcurve* ec;
    curvetype* ctp;
    
    if ( objc != 8 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"type def controls xmin xmax wrapflag savename\"",
		       TCL_STATIC );
	return TCL_ERROR;
    }

    if ( Tcl_GetDoubleFromObj( interp, objv[2], &def ) != TCL_OK )
	return TCL_ERROR;
    if ( Tcl_GetDoubleFromObj( interp, objv[4], &xmin ) != TCL_OK )
	return TCL_ERROR;
    if ( Tcl_GetDoubleFromObj( interp, objv[5], &xmax ) != TCL_OK )
	return TCL_ERROR;
    if ( Tcl_GetBooleanFromObj( interp, objv[6], &wrapflag ) != TCL_OK )
	return TCL_ERROR;

    /* nc is the number of control points. */
    if ( Tcl_ListObjGetElements( interp, objv[3], &nc, &listv ) != TCL_OK )
	return TCL_ERROR;

    /* allocate memory to hold control points. */
    if ( nc > 0 )
    {
	xc = (double*)malloc( nc * sizeof( double ) );
	yc = (double*)malloc( nc * sizeof( double ) );
	tc = (double*)malloc( nc * sizeof( double ) );
	ic = (int*)malloc( nc * sizeof( int ) );
    }
    else
    {
	xc = NULL;
	yc = NULL;
    }

    /* read each control point. */
    for ( i = 0; i < nc; ++i )
    {
	if ( Tcl_ListObjIndex( interp, listv[i], 0, &listel ) != TCL_OK )
	    return TCL_ERROR;
	if ( Tcl_GetDoubleFromObj( interp, listel, xc+i ) != TCL_OK )
	    return TCL_ERROR;

	if ( Tcl_ListObjIndex( interp, listv[i], 1, &listel ) != TCL_OK )
	    return TCL_ERROR;
	if ( Tcl_GetDoubleFromObj( interp, listel, yc+i ) != TCL_OK )
	    return TCL_ERROR;

	ic[i] = i;
    }

    if ( nc > 0 )
    {
	/* sort the control points by x. */

	sortvar = xc;
	qsort( ic, nc, sizeof( int ), index_compare );

	memmove( tc, xc, nc * sizeof( double ) );
	for ( i = 0; i < nc; ++i )
	    xc[i] = tc[ic[i]];
	
	memmove( tc, yc, nc * sizeof( double ) );
	for ( i = 0; i < nc; ++i )
	    yc[i] = tc[ic[i]];

	free( tc );
    }

    /* create an evaluatedcurve structure to hold the
       points returned by the generator function. */
    he = Tcl_CreateHashEntry( &htCurves, (char*)Tcl_GetStringFromObj( objv[7], NULL ), &newhe );
    if ( ! newhe )
    {
	evaluatedcurve* oldec;
	oldec = (evaluatedcurve*)Tcl_GetHashValue( he );
	free( oldec->x );
	free( oldec->y );
	ec = oldec;
    }
    else
    {
	ec = (evaluatedcurve*)malloc( sizeof( evaluatedcurve ) );
    }
    Tcl_SetHashValue( he, ec );
    
    ec->x = NULL;
    ec->y = NULL;
    ec->n = 0;
    ec->def = def;

    if ( nc > 0 )
    {
	/* find the right generator function and call it. */
	for ( ctp = curvetypes; ctp->name != NULL; ++ctp )
	    if ( strcmp ( Tcl_GetStringFromObj( objv[1], NULL ), ctp->name ) == 0 )
	    {
		ctp->generate( xmin, xmax, wrapflag, def, nc, xc, yc, &(ec->n), &(ec->x), &(ec->y) );
		break;
	    }
	free( xc );
	free( yc );
	
	if ( ctp->name == NULL )
	{
	    Tcl_SetResult( interp, "unknown curve type", TCL_STATIC );
	    return TCL_ERROR;
	}
    }

    /* return the number of points generated. */
    Tcl_SetObjResult( interp, Tcl_NewIntObj( ec->n ) );
    
    return TCL_OK;
}

/* GetEvaluatedCurveCmd
 *
 * returns a list of the points describing a curve, so that the curve
 * can be drawn in the edit window.
 */

int GetEvaluatedCurveCmd( ClientData clientdata, Tcl_Interp* interp, 
			  int objc, Tcl_Obj* const objv[] )
{
    Tcl_HashEntry* he;
    evaluatedcurve* ec;
    Tcl_Obj* newlist[2];
    int i;
    
    if ( objc != 2 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"savename\"", TCL_STATIC );
	return TCL_ERROR;
    }

    he = Tcl_FindHashEntry( &htCurves, Tcl_GetStringFromObj( objv[1], NULL ) );
    if ( he == NULL )
    {
	Tcl_SetResult( interp, "no such curve", TCL_STATIC );
	return TCL_ERROR;
    }

    ec = (evaluatedcurve*)Tcl_GetHashValue( he );
    for ( i = 0; i < ec->n; ++i )
    {
	newlist[0] = Tcl_NewDoubleObj( ec->x[i] );
	newlist[1] = Tcl_NewDoubleObj( ec->y[i] );
	Tcl_ListObjAppendElement( interp, Tcl_GetObjResult( interp ),
				  Tcl_NewListObj( 2, newlist ) );
    }

    return TCL_OK;
}

    
    

    
    
