/* -*- compile-command: "msdev modeler.dsp /make \"Modeler - Win32 Debug\"" -*-
 */

#include "modeler.h"

extern "C" {
    int __declspec( dllexport ) Modeler_Init( Tcl_Interp* interp );
    BOOL __declspec( dllexport ) WINAPI DllMain( HINSTANCE, DWORD, LPVOID );

    int Modeler_ResetCmd( ClientData, Tcl_Interp*, int, Tcl_Obj* const [] );
    int Modeler_RebuildCmd( ClientData, Tcl_Interp*, int, Tcl_Obj* const [] );
}

HINSTANCE g_hInstance = NULL;
Tcl_Interp* g_interp = NULL;
Tcl_Obj* g_varDrawmode = NULL;
Tcl_Obj* g_varTimebase = NULL;
Tcl_Obj* g_varTimecurr = NULL;
Tcl_Obj* g_varRayfile = NULL;

int g_quality = 50;
int g_drawmode = 0;

char command_buffer[1024];
char convert_buffer[1024];

FILE* g_rayfile = NULL;

GLfloat mat_ambient[4];
GLfloat mat_diffuse[4];
GLfloat mat_specular[4];
GLfloat mat_shininess;

/* DllMain
 * 
 * every DLL must provide DllMain.  ours just stores the instance handle,
 * so we can use it to create a window later.
 */

BOOL WINAPI DllMain( HINSTANCE hInst, DWORD dwReason, LPVOID reserved )
{
    g_hInstance = hInst;
    return 1;
}

/* Modeler_Init
 *
 * Tcl calls this function when the DLL is first loaded, to initialize
 * the package(s) contained in the DLL.
 */

int Modeler_Init( Tcl_Interp* interp )
{
    Tcl_CreateObjCommand( interp, "::model::reset", Modeler_ResetCmd,
			  NULL, NULL );
    Tcl_CreateObjCommand( interp, "::model::rebuild", Modeler_RebuildCmd,
			  NULL, NULL );

    Curves_Init( interp );
    
    /* initialize the OpenGL window package. */
    if ( OpenGL_Init( interp, g_hInstance ) != TCL_OK )
	return TCL_ERROR;

    /* initialize the UI control package. */
    if ( Controls_Init( interp ) != TCL_OK )
	return TCL_ERROR;

    /* keep a pointer to the interpreter. */
    g_interp = interp;

    /* initialize a Tcl variable with the default
       drawing mode. */
    sprintf( command_buffer, "set g_drawmode %d", MODE_NORMAL | MODE_MED_QUALITY );
    Tcl_GlobalEval( interp, command_buffer );

    /* keep object pointers to Tcl variables we'll want to refer to. */
    g_varDrawmode = Tcl_NewStringObj( "g_drawmode", -1 );
    g_varTimebase = Tcl_NewStringObj( "g_timebase", -1 );
    g_varTimecurr = Tcl_NewStringObj( "g_currtime", -1 );
    g_varRayfile = Tcl_NewStringObj( "g_rayfilename", -1 );
    
    return TCL_OK;
}

/* Modeler_ResetCmd
 *
 * this is called from Tcl whenever a model DLL is unloaded.
 */

int Modeler_ResetCmd( ClientData clientdata, Tcl_Interp* interp,
		      int objc, Tcl_Obj* const objv[] )
{
    /* erase all the control widgets. */
    reset_controls();
    
    return TCL_OK;
}

int Modeler_RebuildCmd( ClientData clientdata, Tcl_Interp* interp,
			int objc, Tcl_Obj* const objv[] )
{
    rebuild_controls();
    
    return TCL_OK;
}

/* get_time
 *
 * models can call this function get the time in seconds since
 * animation mode was last started or reset.
 */

double get_time( void )
{
    double curr;

    /* the time of animation start/reset is stored in the Tcl variable
       "g_timebase".  the time when the model draw command was issued
       is stored in the Tcl variable "g_timecurr".  both variables are
       measured in processor ticks (1000 per second). */
    
    if ( Tcl_GetDoubleFromObj( g_interp,
			       Tcl_ObjGetVar2( g_interp, g_varTimecurr,
					    NULL, TCL_GLOBAL_ONLY ),
			    &curr ) != TCL_OK )
	return 0.0;

    return curr;
}

/* ambient_color
 *
 * models can call this function to set the ambient color for drawing
 * primitives.  this only has an effect if lighting is being used
 * ("normal" drawing mode).
 */

void ambient_color( double r, double g, double b )
{
    mat_ambient[0] = (GLfloat)r;
    mat_ambient[1] = (GLfloat)g;
    mat_ambient[2] = (GLfloat)b;
    mat_ambient[3] = 1.0f;
    
    if ( g_drawmode & MODE_NORMAL )
	glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient );
}

/* diffuse_color
 *
 * models can call this function to set the diffuse color for drawing
 * primitives.  this is also the color used for drawing in wireframe
 * and flat-shaded modes.
 */

void diffuse_color( double r, double g, double b )
{
    mat_diffuse[0] = (GLfloat)r;
    mat_diffuse[1] = (GLfloat)g;
    mat_diffuse[2] = (GLfloat)b;
    mat_diffuse[3] = 1.0f;
	
    if ( g_drawmode & MODE_NORMAL )
	glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse );
    else
	glColor3d( r, g, b );
}

/* specular_color
 *
 * models can call this function to set the specular color for drawing
 * primitives.  this only has an effect if lighting is being used
 * ("normal" drawing mode).
 */

void specular_color( double r, double g, double b )
{
    mat_specular[0] = (GLfloat)r;
    mat_specular[1] = (GLfloat)g;
    mat_specular[2] = (GLfloat)b;
    mat_specular[3] = 1.0f;

    if ( g_drawmode & MODE_NORMAL )
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular );
}

/* shininess
 *
 * models can call this function to set the specular shininess for
 * drawing primitives.  this only has an effect if lighting is being
 * used ("normal" drawing mode).
 */

void shininess( double s )
{
    mat_shininess = (GLfloat)s;
    
    if ( g_drawmode & MODE_NORMAL )
    {
	glMaterialf( GL_FRONT, GL_SHININESS, mat_shininess );
    }
}

/* draw_mode_setup
 *
 * models should call this after clearing the window in their redraw
 * function.  it sets up default material and light properties, and
 * does initialization specific to the drawing mode.
 */

int draw_mode_setup( void )
{
    GLfloat light_position[] = { 10.0f, 10.0f, 10.0f, 0.0f };

    mat_ambient[0] = mat_ambient[1] = mat_ambient[2] = 0.5;
    mat_ambient[3] = 1.0;
    mat_diffuse[0] = mat_diffuse[1] = mat_diffuse[2] = 0.5;
    mat_diffuse[3] = 1.0;
    mat_specular[0] = mat_specular[1] = mat_specular[2] = mat_specular[3] = 1.0;
    mat_shininess = 100.0;
    
    /* the Tcl variable "g_drawmode" contains bit flags controlling
       the drawing mode.  read its value. */
    
    if ( Tcl_GetIntFromObj( g_interp,
			    Tcl_ObjGetVar2( g_interp, g_varDrawmode,
					    NULL, TCL_GLOBAL_ONLY ),
			    &g_drawmode ) != TCL_OK )
	g_drawmode = MODE_NORMAL | MODE_MED_QUALITY;

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    
    glEnable( GL_DEPTH_TEST );

    if ( g_drawmode & MODE_NORMAL )
    {
	/* initialization for "normal" mode. */
	
	glShadeModel( GL_SMOOTH );
	glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_ambient );
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular );
	glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, &mat_shininess );
	glLightfv( GL_LIGHT0, GL_POSITION, light_position );

	glEnable( GL_LIGHTING );
	glEnable( GL_LIGHT0 );
	glEnable( GL_NORMALIZE );
	
	glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE );
	glPolygonMode( GL_FRONT, GL_FILL );
	glCullFace( GL_BACK );
	glEnable( GL_CULL_FACE ); 
    }

    if ( g_drawmode & MODE_WIREFRAME )
    {
	/* initialization for "wireframe" mode. */
	
	glDisable( GL_LIGHTING );
	glDisable( GL_NORMALIZE );
	
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );

	glColor3f( 0.5, 0.5, 0.5 );
    }

    if ( g_drawmode & MODE_FLATSHADE )
    {
	/* initialization for "flat shaded" mode. */
	
	glDisable( GL_LIGHTING );
	glDisable( GL_NORMALIZE );
	
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	
	glColor3f( 0.5, 0.5, 0.5 );
    }
    
    if ( g_drawmode & MODE_RAYTRACE )
    {
	FILE* f;
	Tcl_Obj* fn;

	fn = Tcl_ObjGetVar2( g_interp, g_varRayfile, NULL, TCL_GLOBAL_ONLY );

	Debug( "output file is [%s]\n", Tcl_GetStringFromObj( fn, NULL ) );
	
	f = fopen( Tcl_GetStringFromObj( fn, NULL ), "w" );
	if ( f == NULL )
	{
	    Tcl_SetResult( g_interp, "can't open raytrace output file", TCL_STATIC );
	    return TCL_ERROR;
	}

	fprintf( f, "SBT-raytracer 1.0\n\n" );
	fprintf( f, "camera { fov=30; }\n\n" );
	fprintf( f, "directional_light { direction=(-1,-1,-1); color=(0.7,0.7,0.7); }\n\n" );
		 

	g_rayfile = f;
    }
	    
    
    /* set the quality flag, based on the draw mode. */
    if ( g_drawmode & MODE_HIGH_QUALITY ) g_quality = 50;
    if ( g_drawmode & MODE_MED_QUALITY ) g_quality = 20;
    if ( g_drawmode & MODE_LOW_QUALITY ) g_quality = 10;

    /* tell OpenGL to draw in the back buffer. */
    glDrawBuffer( GL_BACK );

    return TCL_OK;
}

int draw_mode_finish( void )
{
    if ( g_drawmode & MODE_RAYTRACE )
    {
	fclose( g_rayfile );
	g_rayfile = NULL;
    }

    return TCL_OK;
}

/* Debug
 *
 * Debug works just like printf(), except that it prints output in the
 * MSVC++ "output" window.  don't try to print more than 1K bytes at
 * once.
 */

void Debug( char* fmt, ... )
{
    char buffer[1024];
    va_list list;

    va_start( list, fmt );
    vsprintf( buffer, fmt, list );
    va_end( list );

    OutputDebugString( buffer );
}

    
