/* -*- compile-command: "msdev modeler.dsp /make \"Modeler - Win32 Debug\"" -*-
 */

/* OpenGL package
 *
 * this function allows a Win32 window with an OpenGL visual to be
 * created from within Tcl, with Tcl scripts attached for redraw and
 * resize events.
 *
 * it only supports a single opengl window.
 */

#include <math.h>

#include "modeler.h"

static char g_szAppName[] = "Modeler";

static HWND g_hWnd = NULL;
static HINSTANCE g_hInstance = NULL;

/* these Tcl objects should be eval'd when the opengl window needs to
   be refreshed or is resized. */
static Tcl_Obj* g_drawCmd = NULL;
static Tcl_Obj* g_resizeCmd = NULL;

GLsizei g_width, g_height;

/* OpenGL_Init
 *
 * initializes the OpenGL package.
 */

int OpenGL_Init( Tcl_Interp* interp, HINSTANCE hInst )
{
    WNDCLASSEX wndclass;

    g_hInstance = hInst;

    /* register the window class with Win32.  this is all standard,
       except we reserve 4 extra bytes in the window structure
       (cbWndExtra) to keep a copy of the Tcl interpreter pointer,
       which we'll need to eval the draw and resize objects. */
    
    wndclass.cbSize = sizeof( wndclass );
    wndclass.style = CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc = WndProc;
    wndclass.cbClsExtra = 0;
    wndclass.cbWndExtra = 4;
    wndclass.hInstance = g_hInstance;
    wndclass.hIcon = LoadIcon( NULL, IDI_APPLICATION );
    wndclass.hCursor = LoadCursor( NULL, IDC_ARROW );
    wndclass.hbrBackground = (HBRUSH) (COLOR_WINDOW+1);
    wndclass.lpszMenuName = NULL;
    wndclass.lpszClassName = g_szAppName;
    wndclass.hIconSm = LoadIcon( NULL, IDI_APPLICATION );
    RegisterClassEx( &wndclass );

    /* create new commands in the Tcl interpreter for creating,
       destroy, and invalidating (forcing redraw) the opengl
       window. */
    
    Tcl_CreateObjCommand( interp, "::opengl::create", OpenGL_CreateCmd, NULL, NULL );
    Tcl_CreateObjCommand( interp, "::opengl::destroy", OpenGL_DestroyCmd, NULL, NULL );
    Tcl_CreateObjCommand( interp, "::opengl::invalidate", OpenGL_InvalidateCmd, NULL, NULL );

    return TCL_OK;
}

/* OpenGL_InvalidateCmd
 *
 * when this function is called from Tcl, invalidate the output window
 * so that Win32 will ask to redraw it.
 */

int OpenGL_InvalidateCmd( ClientData clientData, Tcl_Interp* interp,
			  int objc, Tcl_Obj* const objv[] )
{
    if ( objc != 1 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"destroy\"",
		       TCL_STATIC );
	return TCL_ERROR;
    }

    if ( g_hWnd )
	InvalidateRect( g_hWnd, NULL, FALSE );
    
    return TCL_OK;
}

/* OpenGL_DestroyCmd
 *
 * when this function is called from Tcl, close the output window.
 */
    
int OpenGL_DestroyCmd( ClientData clientData, Tcl_Interp* interp,
		       int objc, Tcl_Obj* const objv[] )
{
    if ( objc != 1 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"destroy\"",
		       TCL_STATIC );
	return TCL_ERROR;
    }

    if ( g_hWnd )
	SendMessage( g_hWnd, WM_CLOSE, 0, 0 );

    return TCL_OK;
}

/* OpenGL_CreateCmd
 *
 * when this function is called from Tcl, create the output window.
 * it needs two arguments: the script to be evaluated to redraw the
 * window, and the script to handle a window resize.
 */

int OpenGL_CreateCmd( ClientData clientData, Tcl_Interp* interp,
		      int objc, Tcl_Obj* const objv[] )
{
    if ( objc != 3 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"create drawcmd resizecmd\"",
		       TCL_STATIC );
	return TCL_ERROR;
    }

    /* this package only supports a single output window. */
    if ( g_hWnd )
    {
	Tcl_SetResult( interp, "already have vismodel output window",
		       TCL_STATIC );
	return TCL_ERROR;
    }

    /* keep pointers to the draw and resize script objects. */
    g_drawCmd = objv[1];
    Tcl_IncrRefCount( g_drawCmd );
    g_resizeCmd = objv[2];
    Tcl_IncrRefCount( g_resizeCmd );

    /* create a Win32 window.  this is all standard, except that
       OpenGL requires the WS_CLIPCHILDREN and WS_CLIPSIBLINGS
       flags. */
    
    g_hWnd = CreateWindow( g_szAppName,
			   "modeler output", 
			   WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN |
			   WS_CLIPSIBLINGS, 
			   CW_USEDEFAULT, CW_USEDEFAULT,
			   500, 500,
			   NULL, NULL, g_hInstance, NULL );
    /* store a copy of the interpreter pointer in the window
       structure. */
    SetWindowLong( g_hWnd, 0, (LONG)interp );

    /* show the window on the screen. */
    ShowWindow( g_hWnd, SW_SHOWNA );
    UpdateWindow( g_hWnd );

    return TCL_OK;
}

/* WndProc
 *
 * this is called by Win32 to handle events for the output window.
 */

LRESULT CALLBACK WndProc( HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam )
{
    /* a DC is a Win32 "device context".  it is a handle to a Win32
       data structure needed for drawing in a window. */
    static HDC hdc;
    /* an RC is an OpenGL "rendering context".  it is a handle to an
       OpenGL data structure needed for drawing in a window. */
    static HGLRC hrc;
    
    PAINTSTRUCT ps;
    Tcl_Interp* interp;
    Tcl_Obj* temp_obj;
    char temp_string[20];
    
    switch( iMsg )
    {
	case WM_CREATE:
	    /* when the window is created, create an OpenGL RC to go
	       along with it.  we will keep this RC around for the
	       entire lifetime of the window. */
	    hrc = SetupOpenGL( hwnd );
	    return 0;

	case WM_ERASEBKGND:
	    /* we will erase the window ourselves when it is redrawn.
	       tell Win32 NOT to erase the window for us, by
	       returning true. */
	    return 1;

	case WM_PAINT:
	    /* Win32 has requests that the window be redrawn. */

	    /* calling BeginPaint is obligatory for handling a
	       WM_PAINT message.  it returns a DC to let us draw in
	       the window. */
	    hdc = BeginPaint( hwnd, &ps );

	    /* tell OpenGL to use the window's DC and the RC that we
	       created when the window was opened. */
	    wglMakeCurrent( hdc, hrc );

	    /* retrieve a pointer to the Tcl interpreter, which we
	       stashed away in the Win32 window structure. */
	    interp = (Tcl_Interp*)GetWindowLong( hwnd, 0 );

	    /* evaluate the script we were given to redraw the
	       window. */
	    Tcl_GlobalEvalObj( interp, g_drawCmd );

	    /* force OpenGL to finish drawing, then swap buffers so
	       that the newly drawn back buffer appears on the
	       screen. */
	    glFlush();
	    SwapBuffers( hdc );

	    /* tell OpenGL we're done drawing for now. */
	    wglMakeCurrent( NULL, NULL );

	    /* each BeginPaint call must be matched with an EndPaint
	       call to clean up. */
	    EndPaint( hwnd, &ps );

	    /* all done! */
	    return 0;
	
	case WM_SIZE:
	    /* Win32 informs us that the window has been resized. */

	    /* the window's new size is passed to us in lParam. */
	    g_width = (GLsizei)LOWORD( lParam );
	    g_height = (GLsizei)HIWORD( lParam );

	    /* get a DC for the window, and tell OpenGL to use it with
	       the saved RC. */
	    hdc = GetDC( hwnd );
	    wglMakeCurrent( hdc, hrc );

	    /* we create the script to eval by taking the resize
	       script and appending the window's new size.  if the
	       resize script is a procedure name, it will receive the
	       new width and height as arguments. */

	    /* make a copy of the resize script, append the new width
	       and height. */
	    temp_obj = Tcl_DuplicateObj( g_resizeCmd );
	    sprintf( temp_string, " %d %d", g_width, g_height );
	    Tcl_AppendToObj( temp_obj, temp_string, strlen(temp_string) );

	    /* retrieve the interpreter pointer from the window
	       structure, and eval the newly constructed script. */
	    interp = (Tcl_Interp*)GetWindowLong( hwnd, 0 );
	    Tcl_GlobalEvalObj( interp, temp_obj );

	    /* throw away the script we constructed. */
	    Tcl_DecrRefCount( temp_obj );

	    /* tell OpenGL to stop drawing in the window and release
	       the DC. */
	    wglMakeCurrent( NULL, NULL );
	    ReleaseDC( hwnd, hdc );

	    /* all done! */
	    return 0;

	case WM_CLOSE:
	    /* Win32 requests that the window be closed. */

	    /* release our copies of the draw and resize script
	       objects. */
	    if ( g_drawCmd )
	    {
		Tcl_DecrRefCount( g_drawCmd );
		g_drawCmd = NULL;
	    }
	    if ( g_resizeCmd )
	    {
		Tcl_DecrRefCount( g_resizeCmd );
		g_resizeCmd = NULL;
	    }

	    /* delete the window. */
	    DestroyWindow( hwnd );
	    g_hWnd = NULL;
	    
	    return FALSE;
    }

    /* any other messages that we don't handle will be handled by
       DefWindowProc, which does the expected defaulty sort of
       things. */
    return DefWindowProc( hwnd, iMsg, wParam, lParam );
}

/* SetupOpenGL
 *
 * when the output window is created, we must create an OpenGL
 * rendering context (RC) for it.
 */

HGLRC SetupOpenGL( HWND hwnd )
{
    /* the pixel format descriptor tells OpenGL what features we need
       to be supported in the window. */
    static PIXELFORMATDESCRIPTOR pfd = {
	sizeof( PIXELFORMATDESCRIPTOR ),
	1,
	/* request a double-buffered window */
	PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
	PFD_TYPE_RGBA,
	/* request a 24-bit color window */
	24,
	0, 0, 0,
	0, 0, 0,
	0, 0, 
	0, 0, 0, 0, 0,
	/* request a 16-bit depth buffer */
	16,
	0, 
	0,
	PFD_MAIN_PLANE,
	0,
	0,
	0,
	0
    };
    int myPixelFormatID;
    HDC hdc;
    HGLRC hrc;

    /* get a DC for the window. */
    hdc = GetDC( hwnd );

    /* look for a pixel format that matches the requests we put in
       the pixel format descriptor. */
    myPixelFormatID = ChoosePixelFormat( hdc, &pfd );
    if ( myPixelFormatID == 0 )
    {
	/* if no appropriate pixel format could be found, die rather
	   ungracefully. */
	Debug( "this system doesn't support the required visual!" );
	exit( 1 );
    }

    /* set up the window with this pixel format. */
    SetPixelFormat( hdc, myPixelFormatID, &pfd );

    /* create a rendering context for this window. */
    hrc = wglCreateContext( hdc );

    /* release the DC. */
    ReleaseDC( hwnd, hdc );

    return hrc;
}

